#!/usr/bin/env python

import sys
import os

import pygtk
pygtk.require ('2.0')
import gtk
import re
import getopt
import commands
import shutil
import subprocess


HEADING              = "Striata Reader"
SUBMIT               = "Open"
CANCEL               = "Cancel"
USERNAME             = "Username"
PASSWORD             = "Password"
FILE_OPEN_TITLE      = "Open EMC File"
MULTIPLE_MODES       = "Multiple modes specified, please specify only one mode at a time."
DELETE_MESSAGE       = "Click Ok when you have finished viewing the document, so that temporary files may be deleted."
DEFAULT_ERROR        = "An error occurred."
CANNOT_PARSE_OPTIONS = "Could not parse reader options for delete message."
NO_FILE_SELECTED     = "No file selected."
NO_FILE_SPECIFIED    = "No file was specified."
FILE_DOESNT_EXIST    = "Specified file could not be found."
FILE_IS_DIRECTORY    = "Specified path is a directory."
INVALID_FILE         = "Could not read file. Document is possibly corrupt or invalid."

ERROR_CANT_DECODE    = "Unable to decode this document.  You may have entered your user authentication details incorrectly.\nPlease remember that all authentication details are case sensitive."
ERROR_VERSION        = "The chosen file requires a newer version of the Striata Reader in order to be opened\nPlease download the latest version from http://download.striata.com/"
ERROR_CORRUPT        = "Unable to decode this document.  The file appears to be corrupt."
ERROR_HASH           = "Cannot open this document because this version of the Striata Reader\ndoes not support the encryption format used on the selected document.\nPlease download the latest version from http://download.striata.com/"
ERROR_EXPIRED        = "Cannot open this document because the content of this document has expired."
ERROR_WRITE          = "Cannot write files to disk."
ERROR_MEMORY         = "Not enough memory available."

ICON = "/usr/share/striata-reader/icons/72/striata-keymail.png"

READER_BINARY = 'striata-readerc'

KEYMAIL_FLAG_NOPASSWORD   = 0x80000000
KEYMAIL_FLAG_PASSWORDONLY = 0x10000000

KEYMAIL_ERROR_VERSION   = -3
KEYMAIL_ERROR_CORRUPT   = -4
KEYMAIL_ERROR_HASH      = -7
KEYMAIL_ERROR_EXPIRED   = -10
KEYMAIL_ERROR_WRITE     = -13
KEYMAIL_ERROR_MEMORY    = -16

PADDING          = 7
INPUT_SIZE       = 20
INPUT_MAX_LENGTH = 255



# Class OkDialog
class OkDialog:
	def __init__ (self, message, heading = HEADING):
		self.dialog = gtk.Dialog (heading, None, gtk.DIALOG_MODAL, (gtk.STOCK_OK, 1))
		self.dialog.set_position (gtk.WIN_POS_CENTER)
		self.dialog.set_icon_from_file (ICON)

		self.text = gtk.Label (message)
		self.dialog.vbox.pack_start (self.text, True, True, PADDING)
		self.text.show()


	def OkClicked (self, widget, data=None):
		self.dialog.hide()
		self.Quit()
		return True


	def Quit (self, widget, data=None):
		gtk.main_quit()
		return True


	def main (self):
		self.dialog.show()
		return self.dialog.run()



# Class ErrorDialog inherits OkDialog and specifies a default message
class ErrorDialog (OkDialog):
	def __init__ (self, message = DEFAULT_ERROR, heading = HEADING):
		OkDialog.__init__ (self, message, heading)



# Class DeleteDialog inherits OkDialog and is used to determine when the user is done with data
class DeleteDialog (OkDialog):
	def __init__ (self, message = DELETE_MESSAGE, heading = HEADING):
		OkDialog.__init__ (self, message, heading)



# Class FileDialog
class FileDialog:
	def __init__ (self):
		self.fileSelection = gtk.FileSelection (title = FILE_OPEN_TITLE)
		self.fileSelection.set_position (gtk.WIN_POS_CENTER)
		self.fileSelection.connect ("destroy", self.destroyed)

		self.fileSelection.set_icon_from_file (ICON)

	def destroyed (self):
		gtk.main_quit()

	def GetFilename (self):
		return self.fileSelection.get_filename()

	def main (self):
		self.fileSelection.show()
		result = self.fileSelection.run()
		self.fileSelection.hide()
		return result



# Class CredentialDialog
class CredentialDialog:
	def __init__ (self, requireUsername = True, requirePassword = True, prompt = "", heading = HEADING,
			usernameLabel = "", passwordLabel = "", hyperLinkText = "", hyperLinkUrl = "", xdgOpen = ""):
		self.xdgOpen = xdgOpen

		self.window = gtk.Window (gtk.WINDOW_TOPLEVEL)
		self.window.set_position (gtk.WIN_POS_CENTER)
		self.window.connect ("delete_event", self.CancelClicked)
		self.window.set_border_width (10)
		self.window.set_title (heading)

		self.window.set_icon_from_file (ICON)

		self.table = gtk.Table (2, 3)

		if (prompt != ""):
			self.promptLabel = gtk.Label (prompt)
			self.table.attach (self.promptLabel,0,2,0,1,xpadding=PADDING, ypadding=PADDING)
			self.promptLabel.show()

		if (requireUsername):
			self.usernameBox = gtk.HBox (False, 0)
			if (usernameLabel == ""):
				self.usernameLabel = gtk.Label (USERNAME)
			else:
				self.usernameLabel = gtk.Label (usernameLabel)
			self.table.attach (self.usernameLabel, 0, 1, 1, 2, xpadding=PADDING, ypadding=PADDING)
			self.usernameLabel.show()

			self.usernameField = gtk.Entry (INPUT_MAX_LENGTH)
			self.usernameField.set_width_chars (INPUT_SIZE)
			self.table.attach (self.usernameField, 1, 2, 1, 2, xpadding=PADDING, ypadding=PADDING)
			self.usernameField.show();

		if (requirePassword):
			if (passwordLabel == ""):
				self.passwordLabel = gtk.Label (PASSWORD)
			else:
				self.passwordLabel = gtk.Label (passwordLabel)
			self.table.attach (self.passwordLabel, 0, 1, 2, 3, xpadding=PADDING, ypadding=PADDING)
			self.passwordLabel.show()

			self.passwordField = gtk.Entry (INPUT_MAX_LENGTH)
			self.passwordField.set_width_chars (INPUT_SIZE)
			self.passwordField.set_visibility (False)
			self.passwordField.set_activates_default (True)
			self.table.attach (self.passwordField, 1, 2, 2, 3, xpadding=PADDING, ypadding=PADDING)
			self.passwordField.show();

		self.table.show()

		self.buttonBox = gtk.HBox (False, 0)

		self.button = gtk.Button (SUBMIT)
		self.button.connect ("clicked", self.OkClicked, None)
		self.button.set_flags(gtk.CAN_DEFAULT)
		self.window.set_default(self.button)
		self.buttonBox.pack_start (self.button, True, True, PADDING)
		self.button.show()


		self.buttonBox.show()

		self.cancel = gtk.Button (CANCEL)
		self.cancel.connect ("clicked", self.CancelClicked, None)
		self.buttonBox.pack_start (self.cancel, True, True, PADDING)
		self.cancel.show()

		if (hyperLinkText != ""):
			self.hyperLinkUrl = hyperLinkUrl
			self.linkBox = gtk.HBox (False, 0)

			label = gtk.Label ("<span foreground='blue' underline='low'>" + hyperLinkText + '</span>')
			label.set_use_markup (True)
			label.show()

			self.linkButton = gtk.Button()
			self.linkButton.connect ("clicked", self.LinkClicked, None)
			self.linkButton.add (label)
			self.linkButton.set_relief (gtk.RELIEF_NONE)
			self.linkBox.pack_start (self.linkButton)

			#Not available on pygtk < 2.22, but we can safely continue without it.
			try:
				self.linkButton.set_can_focus (False)
			except:
				pass

			self.linkButton.show()
			self.linkBox.show()

		self.mainBox = gtk.VBox (False, 0)
		self.mainBox.pack_start (self.table, True, True, PADDING)
		self.mainBox.pack_start (self.buttonBox, True, True, PADDING)
		if (hyperLinkText != ""):
			self.mainBox.pack_start (self.linkBox, True, True)

		self.window.add (self.mainBox)

		self.mainBox.show()

		# State variables
		self.username = ""
		self.password = ""
		self.requireUsername = requireUsername
		self.requirePassword = requirePassword
		self.okClicked = False


	def CancelClicked (self, widget, data=None):
		return self.Quit(widget,data)


	def OkClicked (self, widget, data=None):
		if (self.requireUsername):
			self.username = self.usernameField.get_text()
		if (self.requirePassword):
			self.password = self.passwordField.get_text()
		self.okClicked = True
		self.window.hide()
		gtk.main_quit()

		return True

	def LinkClicked (self, widget, data=None):
		if self.xdgOpen != "":
			os.system (self.xdgOpen + ' "' + self.hyperLinkUrl + '"')


	def Quit (self, widget, data=None):
		gtk.main_quit()
		return True


	def main (self):
		self.window.show()
		gtk.main()



#Reads emc configuration.
class OptionParser:
	def __init__ (self):
		self.usernameRequired = False
		self.passwordRequired = False
		self.keyValues = {}

	def parse (self, text=""):
		matcher = re.compile ("(\w+)\<\=\>(.*)")
		key     = ""
		lastKey = ""
		lines   = []

		lines = text.split ("\n");
		for line in lines:
			match = matcher.match (line)
			if match is None:
				text = text + "\n" + line
			else:
				if (key != ""):
					#Strip the extra newline if it exists:
					if text[-1:] == '\n':
						text = text[:-1]
					self.keyValues[key] = text

				lastKey = key
				key = match.group(1)
				text = match.group(2)

		# slurp up last line
		if (lastKey != key):
			if text[-1:] == '\n':
				text = text[:-1]
			self.keyValues[key] = text

		#Decode the username and password flags
		flags = self.keyValues['flags']

		if ((int (flags) & (KEYMAIL_FLAG_NOPASSWORD)) or (int (flags) & (KEYMAIL_FLAG_PASSWORDONLY))):
			self.usernameRequired = False
		else:
			self.usernameRequired = True

		if (KEYMAIL_FLAG_NOPASSWORD & int (flags)):
			self.passwordRequired = False
		else:
			self.passwordRequired = True



class Main:
	def __init__ (self):
		errorMsg = ""
		fileName = ""
		self.xdgOpen = ""
		self.username = ""
		self.password = ""
		self.tempDir = ""

		if (not (os.environ.has_key ("DISPLAY"))):
			sys.exit (1)

		options = getopt.getopt (sys.argv[1:], "f:t:x:")

		for option in options[0]:
			if (option[0] == '-f'):
				fileName = option[1]

			elif (option[0] == '-t'):
				self.tempDir = option[1]

			elif (option[0] == '-x'):
				self.xdgOpen = option[1]

		self.RunReader (fileName)

		sys.exit (0)


	def RunCredentials(self):
		optionParser = OptionParser()
		optionParser.parse (self.strings)

		title = optionParser.keyValues["title"]
		prompt = optionParser.keyValues["prompt"]
		usernameLabel = optionParser.keyValues["userNameLabel"]
		passwordLabel = optionParser.keyValues["passwordLabel"]
		hyperLinkUrl = optionParser.keyValues["hyperlinkUrl"]
		hyperLinkText = optionParser.keyValues["hyperlinkText"]
		if (optionParser.usernameRequired != True and optionParser.passwordRequired != True):
			self.username = ""
			self.password = ""
			return (0)

		dialog = CredentialDialog (optionParser.usernameRequired, optionParser.passwordRequired,
			prompt, title, usernameLabel, passwordLabel, hyperLinkText, hyperLinkUrl, self.xdgOpen)
		dialog.main()

		if (dialog.okClicked):
			self.username = dialog.username
			self.password = dialog.password
			return (0)
		else:
			sys.exit (0)


	def MultipleModes (self):
		self.Error (MULTIPLE_MODES)



	def RunReader (self, fileName):
		#Make sure we have a valid fileName
		while (fileName == ""):
			fileSelect = FileDialog()
			result = fileSelect.main()

			if (result == -4 or result == -6):
				sys.exit (6)

			if (result != -5):
				self.Error (NO_FILE_SELECTED)

			fileName = fileSelect.GetFilename()
			if (not os.path.isfile (fileName)):
				if (os.path.exists (fileName)):
					dialog = ErrorDialog (FILE_IS_DIRECTORY)
					dialog.main()
				else:
					dialog = ErrorDialog (FILE_DOESNT_EXIST)
					dialog.main()

				fileName = ""

		if (not os.path.isfile (fileName)):
			if (os.path.exists (fileName)):
				dialog = ErrorDialog (FILE_IS_DIRECTORY)
				dialog.main()
				self.Error (FILE_IS_DIRECTORY)
			else:
				dialog = ErrorDialog (FILE_DOESNT_EXIST)
				dialog.main()
				self.Error (FILE_DOESNT_EXIST)

		# Ensure we have a valid EMC document, and capture strings
		(status, strings) = commands.getstatusoutput (READER_BINARY + ' -internalguicall "' + fileName + '"')
		status = self.CalculateStatus (status)

		if status != 0:
			self.ShowDecodingError (status, None, fileName)
			sys.exit (5)

		self.strings = strings

		try:
			os.makedirs (self.tempDir)
			os.chdir (self.tempDir)
		except OSError:
			self.ShowDecodingError (KEYMAIL_ERROR_WRITE, optionParser, fileName)
			sys.exit (1)

		#Get credentials.
		self.RunCredentials()

		optionParser = OptionParser()
		optionParser.parse (self.strings)

		# Setup arguments.
		arguments = [READER_BINARY]
		if self.xdgOpen != "":
			arguments.append ('-openwith=' + self.xdgOpen)

		if optionParser.usernameRequired == True:
			arguments.append ('-username=' + self.username)

		if optionParser.passwordRequired == True:
			arguments.append ('-password=' + self.password)

		arguments.append (fileName)
		os.umask (077)

		DEVNULL = open (os.devnull, 'w')
		status = subprocess.call (arguments, stdout=DEVNULL, stderr=DEVNULL)

		if status < 0:
			self.ShowDecodingError (status, optionParser, fileName)

		elif status == 0:
			message = ""
			if self.xdgOpen != "":
				message = optionParser.keyValues["okToDelete"]
			else:
				message = "Document has been extracted to " + self.tempDir + " delete files when done."

			heading = optionParser.keyValues["okToDeleteTitle"]
			dialog = OkDialog (message, heading)
			dialog.main()
		else:
			message = optionParser.keyValues["cantDecodeFile"]
			heading = optionParser.keyValues["errorTitle"]
			dialog = OkDialog (message, heading)
			dialog.main()

		if self.xdgOpen != "":
			self.Cleanup()


	def ShowDecodingError (self, errorCode, optionParser, fileName):
		errorMsg = ""
		heading = ""

		if optionParser is not None:
			heading = optionParser.keyValues["errorTitle"]

		if heading == "":
			heading = HEADING

		(status, output) = commands.getstatusoutput (READER_BINARY + ' -geterrorstring=' + str (errorCode) + ' "' + fileName + '"')
		status = self.CalculateStatus (status)
		if status == 0:
			errorMsg = output

		elif errorCode == -241:
			if optionParser is not None:
				errorMsg = optionParser.keyValues["cantDecodeFile"]
			if errorMsg == "":
				errorMsg = ERROR_CANT_DECODE

		elif errorCode == KEYMAIL_ERROR_MEMORY:
			if optionParser is not None:
				errorMsg = optionParser.keyValues["outOfMemory"]
			if errorMsg == "":
				errorMsg = ERROR_MEMORY

		elif errorCode == KEYMAIL_ERROR_WRITE:
			if optionParser is not None:
				errorMsg = optionParser.keyValues["cantCreateTempFile"]
			if errorMsg == "":
				errorMsg = ERROR_WRITE

		elif errorCode == KEYMAIL_ERROR_EXPIRED:
			if optionParser is not None:
				errorMsg = optionParser.keyValues["expired"]
			if errorMsg == "":
				errorMsg = ERROR_EXPIRED

		elif errorCode == KEYMAIL_ERROR_CORRUPT:
			if optionParser is not None:
				errorMsg = optionParser.keyValues["fileCorrupt"]
			if errorMsg == "":
				errorMsg = ERROR_CORRUPT

		elif errorCode == KEYMAIL_ERROR_VERSION:
			if optionParser is not None:
				errorMsg = optionParser.keyValues["incorrectVersion"]
			if errorMsg == "":
				errorMsg = ERROR_VERSION

		elif errorCode == KEYMAIL_ERROR_HASH:
			if optionParser is not None:
				errorMsg = optionParser.keyValues["notSupported"]
			if errorMsg == "":
				errorMsg = ERROR_HASH

		if errorMsg == "":
			errorMsg = "Could not decode file."

		dialog = ErrorDialog (errorMsg, heading)
		dialog.main()


	def CalculateStatus (self, value):
		return (value >> 8) & 0x00FF


	def Cleanup (self):
		if os.path.isdir (self.tempDir):
			shutil.rmtree (self.tempDir)


	def Error (self, message):
		print "Error: " + message
		sys.exit(5)


# Main entry point
if __name__ == "__main__":
	Main()
